import os

from django.conf import settings, global_settings


if not settings.configured:
    # Helper function to extract absolute path
    location = lambda x: os.path.join(os.path.dirname(os.path.realpath(__file__)), x)

    settings.configure(
            DATABASES={
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': ':memory:',
                }
            },
            INSTALLED_APPS=[
                'django.contrib.auth',
                'django.contrib.admin',
                'django.contrib.contenttypes',
                'django.contrib.sessions',
                'django.contrib.sites',
                'django.contrib.flatpages',
                'url_tracker',
            ],
            TEMPLATE_CONTEXT_PROCESSORS=(
                "django.contrib.auth.context_processors.auth",
                "django.core.context_processors.request",
                "django.core.context_processors.debug",
                "django.core.context_processors.i18n",
                "django.core.context_processors.media",
                "django.core.context_processors.static",
                "django.contrib.messages.context_processors.messages",
            ),
            TEMPLATE_DIRS=(
                location('templates'),
            ),
            MIDDLEWARE_CLASSES=global_settings.MIDDLEWARE_CLASSES + (
                'url_tracker.middleware.URLChangePermanentRedirectMiddleware',
            ),
            AUTHENTICATION_BACKENDS=(
                'django.contrib.auth.backends.ModelBackend',
            ),
            ROOT_URLCONF='tests.urls',
            DEBUG=False,
            SITE_ID=1,
            APPEND_SLASH=True,
        )



"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""
from mock import Mock

from django.http import Http404
from django.test import TestCase

import url_tracker
from url_tracker.models import URLChangeRecord


class TestTracking(TestCase):

    def setUp(self):
        class DoesNotExist(BaseException):
            pass

        self.model_mock = Mock
        self.model_mock.get_absolute_url = Mock(method='get_absolute_url')

        self.tracked_model = self.model_mock(name='TrackedModel')
        self.tracked_model._get_tracked_url = lambda: u'/the/new/one/' 

        def raise_exception(*args, **kwargs):
            raise self.tracked_model.__class__.DoesNotExist

        class_objects = Mock(name="MockModelManager")
        class_objects.get = raise_exception

        self.tracked_model.__class__.objects = class_objects
        self.tracked_model.__class__.DoesNotExist = DoesNotExist

        self.tracked_db_model = self.model_mock(name='TrackeDatabaseModel')
        self.tracked_db_model._get_tracked_url = lambda: u'/the/old/one/'

    def test_tracking_model_without_url_method(self):

        class EmptyModel(object):
            pass

        self.assertRaises(
            url_tracker.URLTrackingError,
            url_tracker.track_url_changes_for_model,
            EmptyModel(),
        )

    def test__lookup_url_with_new_instance(self):
        url_tracker.track_url_changes_for_model(Mock)
        url_tracker.lookup_previous_url(self.tracked_model)

        self.assertEquals(self.tracked_model._old_url, None)

    def test_lookup_url_with_existing_instance(self):
        def return_instance(pk):
            return self.tracked_db_model

        class_objects = Mock(name='MockModelManager')
        class_objects.get = return_instance
        self.tracked_model.__class__.objects = class_objects

        url_tracker.track_url_changes_for_model(Mock)
        url_tracker.lookup_previous_url(self.tracked_model)

        self.assertEquals(self.tracked_model._old_url, u'/the/old/one/')

    def test_track_changed_url_with_new_instance(self):
        instance = self.tracked_model
        instance._old_url = None

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 0)

    def test_track_changed_url_with_unchanged_url(self):
        instance = self.tracked_model
        instance._old_url = '/the/new/one/' 

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 0)

    def test_track_changed_url_without_existing_records(self):
        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 1)
        record = URLChangeRecord.objects.all()[0]
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.old_url, u'/the/old/one/')
        self.assertEquals(record.deleted, False)

    def test_track_changed_url_with_existing_records(self):
        URLChangeRecord.objects.create(old_url='/the/oldest/one/', new_url='/the/old/one/')
        URLChangeRecord.objects.create(old_url='/one/', new_url='/the/')

        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 3)
        record = URLChangeRecord.objects.get(pk=1)
        self.assertEquals(record.old_url, u'/the/oldest/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)
        record = URLChangeRecord.objects.get(pk=3)
        self.assertEquals(record.old_url, u'/the/old/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)

    def test_track_changed_url_with_existing_records_and_old_url(self):
        URLChangeRecord.objects.create(old_url='/the/oldest/one/', new_url='/the/old/one/')
        URLChangeRecord.objects.create(old_url='/the/old/one/', new_url='/the/')

        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 2)
        record = URLChangeRecord.objects.get(pk=1)
        self.assertEquals(record.old_url, u'/the/oldest/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)
        record = URLChangeRecord.objects.get(pk=2)
        self.assertEquals(record.old_url, u'/the/old/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)

    def test_track_changed_url_with_existing_deleted_record(self):
        URLChangeRecord.objects.create(old_url='/the/oldest/one/',
                                       new_url='/the/old/one/',
                                       deleted=True)
        URLChangeRecord.objects.create(old_url='/one/', new_url='/the/')

        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_changed_url(instance)

        record = URLChangeRecord.objects.get(pk=3)
        self.assertEquals(record.old_url, u'/the/old/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)

    def test_track_deleted_url_without_existing_records(self):
        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_deleted_url(instance)

        self.assertEquals(URLChangeRecord.objects.count(), 1)
        record = URLChangeRecord.objects.all()[0]
        self.assertEquals(record.new_url, None)
        self.assertEquals(record.old_url, '/the/old/one/')
        self.assertEquals(record.deleted, True)

    def test_track_changed_url_deleting_exsiting_record_with_new_url(self):
        URLChangeRecord.objects.create(old_url='/the/new/one/', new_url='/the/')

        instance = self.tracked_model
        instance._old_url = '/the/old/one/'

        url_tracker.track_changed_url(instance)
        self.assertEquals(URLChangeRecord.objects.count(), 1)
        record = URLChangeRecord.objects.get(pk=1)
        self.assertEquals(record.old_url, u'/the/old/one/')
        self.assertEquals(record.new_url, u'/the/new/one/')
        self.assertEquals(record.deleted, False)


class TestUrlChanges(TestCase):

    def test_invalid_url(self):
        response = self.client.get('/work/an-invalid-project/')
        self.assertEquals(response.status_code, 404)

    def test_changed_url(self):
        url_change = URLChangeRecord.objects.create(
            old_url='/the/old-url/',
            new_url='/the/new/url/',
        )

        response = self.client.get('/the/old-url/')
        self.assertEquals(response.status_code, 301)
        self.assertEquals(response['location'], 'http://testserver/the/new/url/')

    def test_deleted_url(self):
        url_change = URLChangeRecord.objects.create(
            old_url='/the/old-url/',
            new_url='',
            deleted=True
        )

        response = self.client.get('/the/old-url/')
        self.assertEquals(response.status_code, 410)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'URLChange'
        db.create_table('url_tracker_urlchange', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('old_url', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('new_url', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('deleted', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('date_changed', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
        ))
        db.send_create_signal('url_tracker', ['URLChange'])


    def backwards(self, orm):
        
        # Deleting model 'URLChange'
        db.delete_table('url_tracker_urlchange')


    models = {
        'url_tracker.urlchange': {
            'Meta': {'object_name': 'URLChange'},
            'date_changed': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'deleted': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'new_url': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'old_url': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['url_tracker']



import logging
logger = logging.getLogger(__file__)

from django.db.models import signals

from url_tracker.models import URLChangeRecord


class URLTrackingError(Exception):
    """
    Exception raised when an error occures during URL tracking.
    """
    pass


def lookup_previous_url(instance, **kwargs):
    """
    Look up the absolute URL of *instance* from the database while it is
    in a ``pre_save`` state. The previous url is saved in the instance as
    *_old_url* so that it can be used after the instance was saved. 

    If the instance has not been saved to the database (i.e. is new) the
    *_old_url* will be stored as ``None`` which will prevent further tracking
    after saving the instance.
    """
    try:
        db_instance = instance.__class__.objects.get(pk=instance.pk)
        logger.debug("saving old URL for instance '%s' URL", instance.__class__.__name__)
        instance._old_url = db_instance._get_tracked_url()
    except instance.__class__.DoesNotExist:
        logger.debug("new instance, no URL tracking required")
        instance._old_url = None


def track_changed_url(instance, **kwargs):
    """
    Track a URL change for *instance* after a new instance was saved. If 
    the old URL is ``None`` (i.e. *instance* is new) or the new URL and
    the old one are equal (i.e. URL is unchanged), nothing will be changed
    in the database. 

    For URL changes, the database will be checked for existing records that
    have a *new_url* entry equal to the old URL of *instance* and updates
    these records. Then, a new ``URLChangeRecord`` is created for the
    *instance*.
    """
    old_url = getattr(instance, '_old_url', None)
    
    if old_url is None:
        return

    new_url = instance._get_tracked_url()
    # we don't want to store URL changes for unchanged URL
    if old_url == new_url:
        return

    logger.debug(
        "tracking URL change for instance '%s' URL",
        instance.__class__.__name__
    )
    
    # check if the new URL is already in the table and 
    # remove these entries
    for record in URLChangeRecord.objects.filter(old_url=new_url):
        record.delete()

    # updated existing records with the old URL being
    # the new URL in the record
    url_records = URLChangeRecord.objects.filter(new_url=old_url)
    for record in url_records:
        record.new_url = new_url
        record.deleted = False
        record.save()

    # create a new/updated record for this combination of old and
    # new URL. If the record already exists, it is assumed that the
    # current change is to be used and the existing new_url will be
    # detached from the old_url.
    try:
        logger.info(
            "a record for '%s' already exists and will be updated",
            instance._old_url,
        )
        record = URLChangeRecord.objects.get(old_url=instance._old_url)
        record.new_url = new_url
        record.deleted = False
        record.save()
    except URLChangeRecord.DoesNotExist:
        URLChangeRecord.objects.create(
            old_url=instance._old_url,
            new_url=new_url,
            deleted=False
        )


def track_deleted_url(instance, **kwargs):
    """
    Track the URL of a deleted *instance*. It updates all existing 
    records with ``new_url`` being set to the *instance*'s old URL and
    marks this record as deleted URL. 
    
    A new ``URLChangeRecord`` is created for the old URL of *instance*
    that is marked as deleted.
    """
    logger.debug("tracking deleted instance '%s' URL", instance.__class__.__name__)
    old_url = getattr(instance, '_old_url', None)
    if old_url is None:
        return

    # updated existing records with the old URL being the new_url
    # of this record. Changed the *deleted* flag to be ``False``
    url_records = URLChangeRecord.objects.filter(new_url=old_url)
    for record in url_records:
        record.new_url = ''
        record.deleted = True
        record.save()

    try:
        url_change = URLChangeRecord.objects.get(old_url=old_url)
        url_change.deleted = True
        url_change.save()
    except URLChangeRecord.DoesNotExist:
        urlchange = URLChangeRecord.objects.create(
            old_url=old_url,
            deleted=True
        )


def track_url_changes_for_model(model, absolute_url_method='get_absolute_url'):
    """
    Keep track of URL changes of the specified *model*. This will connect the 
    *model*'s ``pre_save``, ``post_save`` and ``post_delete`` signals to the 
    tracking methods ``lookup_previous_url``, ``track_changed_url`` 
    and ``track_deleted_url`` respectively. URL changes will be logged in the
    ``URLChangeRecord`` table and are handled by the tracking middleware when 
    a changed URL is called.
    """
    try:
        model._get_tracked_url = getattr(model, absolute_url_method)
    except AttributeError:
        raise URLTrackingError(
            "cannot track instance %s without method %s",
            model.__class__.__name__,
            absolute_url_method,
        )

    signals.pre_save.connect(lookup_previous_url, sender=model, weak=False)
    signals.post_save.connect(track_changed_url, sender=model, weak=False)
    signals.post_delete.connect(track_deleted_url, sender=model, weak=False)



from django import http
from django.conf import settings

from url_tracker.models import URLChangeRecord


class URLChangePermanentRedirectMiddleware(object):

    def process_response(self, request, response):
        if response.status_code != 404:
            return response # No need to check for non-404 responses.
        try:
            requested_url = request.path_info
            redirect_url = URLChangeRecord.objects.get(old_url__exact=requested_url)
            if redirect_url.deleted:
                return http.HttpResponseGone()
            else:
                return http.HttpResponsePermanentRedirect(redirect_url.new_url)
        # Return the original response if any errors happened. Because this
        # is a middleware, we can't assume the errors will be caught elsewhere.
        except (http.Http404, URLChangeRecord.DoesNotExist):
            return response
        except:
            if settings.DEBUG:
                raise
            return response
